#pragma once

#include "alibxx.hpp"
#include <stdint.h>


namespace A_LIB_NAMESPACE{

namespace Type{


// signed type

typedef int8_t Byte;

typedef int16_t Int16;

typedef int32_t Int32;

typedef int64_t Int64;

// unsigned bytes

typedef u_int8_t UByte;

typedef u_int16_t UInt16;

typedef u_int32_t UInt32;

typedef u_int64_t UInt64;





} // Type


} // Alibxx Namespace
